/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.component.framework.listener;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.RejectedExecutionException;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.LogManager;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.messaging.MessageExchange;

import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.ComponentConfiguration;
import org.ow2.petals.component.framework.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.util.StringHelper;

/**
 * This class helps a Component to listen to {@link MessageExchange}s on the
 * DeliveryChannel (accept() method). It acts as a ListenerManager. It starts or
 * stops a pool of thread that listen to new incoming JBI
 * {@link MessageExchange}s<p/>
 * <p>
 * 
 * @version $Revision: $ $Date: $
 * @since Petals 1.0
 * @author alouis, ofabre
 */
public class JBIListenerManager {

    private Logger logger; // NOPMD by gblondelle

    private LinkedBlockingQueue<Runnable> workers;

    private ThreadPoolExecutor executor;

    private ComponentConfiguration configuration;

    private AbstractComponent component;

    private Map<String, MessageExchangeSynchronizer> asynchronousMessages;

    /**
     * Creates a new instance of {@link JBIListenerManager}
     * 
     * @param channel
     * @param configuration
     *            The configuration of component
     * @param componentName
     *            Name of the component. Used to define the Thread name.
     * @param log
     * @param suManager
     * @param listener
     */
    public JBIListenerManager(AbstractComponent component,
            final ComponentConfiguration configuration, Logger log) {
        super();
        this.component = component;
        this.configuration = configuration;
        if (log != null) {
            this.logger = log;
        } else {
            this.logger = LogManager.getLogManager().getLogger(
                    component.getContext().getComponentName());
        }
        this.workers = new LinkedBlockingQueue<Runnable>(configuration.getPoolSize());
        this.asynchronousMessages = Collections
                .synchronizedMap(new HashMap<String, MessageExchangeSynchronizer>());
        cleanMESynchronizers();
    }

    /**
     * Start threads that listen to {@link MessageExchange}s
     * 
     * @throws JBIException
     * 
     */
    public void startListening() throws JBIException {
        this.executor = new ThreadPoolExecutor(configuration.getPoolSize(), configuration
                .getPoolSize(), Long.MAX_VALUE, TimeUnit.SECONDS, workers,
                new JBIListenerThreadFactory(component.getContext().getComponentName()));
        for (int i = 0; i < configuration.getPoolSize(); i++) {
            try {
                MessageExchangeProcessor processor = new MessageExchangeProcessor(this.component);
                AbstractJBIListener listener = createAndInitJBIListener();
                processor.setJBIListener(listener);
                processor.setLogger(logger);
                executor.execute(processor);
            } catch (RejectedExecutionException e) {
                // TODO Treat this exception
                logger.log(Level.SEVERE, e.getMessage(), e);
            } catch (PEtALSCDKException e) {
                throw new JBIException("Cannot create JBIListener !", e);
            }
        }
    }

    /**
     * Instantiate and initialize a JBI listener.
     * 
     * @return
     * @throws PEtALSCDKException
     */
    private AbstractJBIListener createAndInitJBIListener() throws PEtALSCDKException {
        AbstractJBIListener jbiListener = null;
        String jbiListenerClassName = configuration.getJbiListenerClassName();
        this.logger.log(Level.FINE, "Creating JBIListener '" + jbiListenerClassName + "'");

        if (!StringHelper.isNullOrEmpty(jbiListenerClassName)) {
            try {
                ClassLoader componentClassLoader = this.getClass().getClassLoader();
                Class<?> clazz = componentClassLoader.loadClass(jbiListenerClassName);
                jbiListener = (AbstractJBIListener) clazz.newInstance();
            } catch (InstantiationException e) {
                throw new PEtALSCDKException("Can't instanciate JBIListener : "
                        + jbiListenerClassName, e);
            } catch (IllegalAccessException e) {
                throw new PEtALSCDKException("Can't instanciate JBIListener : "
                        + jbiListenerClassName, e);
            } catch (ClassNotFoundException e) {
                throw new PEtALSCDKException("Can't find JBIListener class : "
                        + jbiListenerClassName, e);
            }
            jbiListener.init(component);
            jbiListener.init();
        } else {
            throw new PEtALSCDKException("JBIListener class name can not be empty");
        }
        return jbiListener;
    }

    /**
     * Stop threads that listen to {@link MessageExchange}s
     * 
     */
    public void stopListening() {
        this.executor.shutdownNow();
        try {
            // arbitrary duration of the wait, as the thread running this method
            // has alreaday a timeout
            this.executor.awaitTermination(60, TimeUnit.SECONDS);
        } catch (InterruptedException e) {
            // do nothing, the thread processing this method has been
            // interrupted due to the reach of its timeout
        }
    }

    /**
     * Add an asynchronous message to the asynchronous messages map.
     * 
     * @param id
     * @param exchangeSynchronizer
     */
    public void addAsynchronousMessage(String id, MessageExchangeSynchronizer exchangeSynchronizer) {
        synchronized (this.asynchronousMessages) {
            logger.finest("Add synchronizer for exchange id: " + id);
            this.asynchronousMessages.put(id, exchangeSynchronizer);
        }
    }

    /**
     * Get an asynchronous message from its id.
     * 
     * @param id
     * @return null if the asynchronous message is not found in the map
     */
    public MessageExchangeSynchronizer getAsynchronousMessage(String id) {
        return asynchronousMessages.get(id);
    }

    /**
     * Remove an asynchronous message from the map
     * 
     * @param id
     */
    public void removeAsynchronousMessage(String id) {
        synchronized (this.asynchronousMessages) {
            logger.finest("Remove synchronizer for exchange id: " + id);
            this.asynchronousMessages.remove(id);
        }
    }

    /**
     * Periodically clean ME synchronizer map. A synchronizer can be removed
     * only if ME response has been accepted and provider has been acknowledged
     */
    private void cleanMESynchronizers() {
        Timer timer = new Timer();
        MESynchronizerCleaner cleaner = new MESynchronizerCleaner();
        timer.schedule(cleaner, 0, 2000);
    }

    public class MESynchronizerCleaner extends TimerTask {
        @Override
        public void run() {
            ArrayList<MessageExchangeSynchronizer> synchronizers = new ArrayList<MessageExchangeSynchronizer>(
                    asynchronousMessages.values());
            for (MessageExchangeSynchronizer synchronizer : synchronizers) {
                if (synchronizer.isAcknowledged()) {
                    removeAsynchronousMessage(synchronizer.getExchange().getExchangeId());
                }
            }
        }
    }
}
